#########################

use Test::More tests => 15;

BEGIN { use_ok('PTools::SDF') };
BEGIN { use_ok('PTools::SDF::ARRAY') };
BEGIN { use_ok('PTools::SDF::CSV') };
BEGIN { use_ok('PTools::SDF::DIR') };
BEGIN { use_ok('PTools::SDF::File') };
BEGIN { use_ok('PTools::SDF::IDX') };
BEGIN { use_ok('PTools::SDF::INI') };
BEGIN { use_ok('PTools::SDF::Lock::Advisory') };
BEGIN { use_ok('PTools::SDF::Lock::Selective') };
BEGIN { use_ok('PTools::SDF::SDF') };
BEGIN { use_ok('PTools::SDF::Sort::Bubble') };
BEGIN { use_ok('PTools::SDF::Sort::Quick') };
BEGIN { use_ok('PTools::SDF::Sort::Random') };
BEGIN { use_ok('PTools::SDF::Sort::Shell') };
BEGIN { use_ok('PTools::SDF::TAG') };
#########################
